from .api import load_data, load_model, metric, preprocess, slicer, transform
from .runner import __create_parser  # type: ignore

__all__ = [
    "load_data",
    "load_model",
    "preprocess",
    "slicer",
    "transform",
    "metric",
    "__create_parser",
]



import functools
from typing import Tuple, Union


def load_model(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.load_model = True
    return _wrapper


def load_data(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.load_data = True
    return _wrapper


def preprocess(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.preprocess = True
    return _wrapper


def slicer(metrics: Union[str, Tuple[str, str]]):
    """Decorator for slicing functions.

    Args:
        metrics: A list of metrics or [tranform, metric] tuples.


    Example::

        @slicer(["accuracy"], ["rotate", "false_positive"])
        def small_sample(data, metadata):
            return data.sample(10), metadata
    """

    def _decorate(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        _wrapper.slicer = True
        _wrapper.metrics = metrics
        return _wrapper

    return _decorate


def transform(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.transform = True
    return _wrapper


def metric(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.metric = True
    return _wrapper



import argparse
from multiprocessing import Pipe, Process
from pathlib import Path

from .server import run_background_processor, run_server


def __create_parser():
    parser = argparse.ArgumentParser(description="Evaluate ML systems.")
    group_a = parser.add_argument_group("Core test options")
    group_b = parser.add_argument_group("Additional options")
    group_a.add_argument(
        "test-files",
        metavar="test_files",
        nargs="+",
        type=Path,
        help="Directory or Python files with annotated functions"
        + "such as slicers and metrics.",
    )
    group_a.add_argument(
        "--metadata", nargs=1, type=Path, help="CSV with metadata for each instance."
    )
    group_a.add_argument(
        "--data-path",
        dest="data_path",
        nargs="?",
        type=Path,
        default="",
        help="Folder with data instances identified"
        + "by the id-column option in the metadata file.",
    )
    group_a.add_argument(
        "--models",
        dest="models",
        type=Path,
        nargs="+",
        help="Paths to models for testing",
    )
    group_b.add_argument(
        "--id-column",
        dest="id_column",
        default="id",
        type=str,
        nargs="?",
        help="Column with the ID used retrieve data files AND for caching",
    )
    group_b.add_argument(
        "--label-column",
        dest="label_column",
        default="label",
        type=str,
        nargs="?",
        help="Column with the ground truth label for instances.",
    )
    group_b.add_argument(
        "--batch-size",
        dest="batch_size",
        nargs="?",
        default=64,
        type=int,
        help="Batch size of passed model predictions",
    )
    group_b.add_argument(
        "--cache-path",
        dest="cache_path",
        nargs="?",
        type=Path,
        default="./.zeno_cache/",
        help="Folder for caching results of slicers and metrics.",
    )
    return parser


def main():
    parser = __create_parser()
    args = parser.parse_args()

    server_conn, processor_conn = Pipe()

    server_process = Process(target=run_server, args=(server_conn, args))
    background_process = Process(
        target=run_background_processor,
        args=(
            processor_conn,
            args,
        ),
    )

    server_process.start()
    background_process.start()

    server_process.join()



import asyncio
import json
import os

import uvicorn  # type: ignore
from fastapi import FastAPI, WebSocket
from fastapi.responses import Response
from fastapi.staticfiles import StaticFiles

from .zeno import Zeno


def run_background_processor(conn, args):
    zeno = Zeno(
        metadata_path=args.metadata[0],
        test_files=args.__dict__["test-files"],
        models=args.models,
        batch_size=args.batch_size,
        id_column=args.id_column,
        label_column=args.label_column,
        data_path=args.data_path,
        cache_path=args.cache_path,
    )

    zeno.start_processing()

    while True:
        case, options = conn.recv()

        if case == "GET_SLICERS":
            slicers = zeno.get_slicers()
            conn.send(
                json.dumps(
                    [
                        {
                            "name": s.name_list,
                            "source": s.source,
                            "slices": s.slices,
                        }
                        for s in slicers
                    ]
                )
            )

        if case == "GET_METRICS":
            testers = zeno.get_metrics()
            conn.send(
                json.dumps([{"name": s.name, "source": s.source} for s in testers])
            )

        if case == "GET_SLICES":
            slices = zeno.get_slices()
            conn.send(
                json.dumps(
                    [
                        {
                            "name": s.name,
                            "size": s.size,
                            "id_column": zeno.id_column,
                            "label_column": zeno.label_column,
                        }
                        for s in slices
                    ]
                )
            )

        # if case == "GET_DATA":
        #     conn.send(zeno.get_metadata_bytes())

        if case == "GET_TABLE":
            conn.send(zeno.get_table(options))

        if case == "GET_MODEL_OUTPUTS":
            conn.send(json.dumps(zeno.get_model_outputs(options)))

        if case == "GET_RESULTS":
            res = zeno.get_results()
            res = [
                {
                    "id": hash(r),
                    "metric": r.metric,
                    "transform": r.transform,
                    "slice": r.sli,
                    "sliceSize": r.slice_size,
                    "modelResults": r.model_results,
                    "modelNames": [str(n) for n in r.model_names],
                    "modelIds": list(r.model_results.keys()),
                }
                for r in res
            ]
            conn.send((zeno.get_status(), res))


def run_server(conn, args):
    app = FastAPI(title="Frontend API")
    api_app = FastAPI(title="Backend API")

    if args.data_path != "":
        app.mount("/static", StaticFiles(directory=args.data_path), name="static")
    app.mount("/api", api_app)
    app.mount(
        "/",
        StaticFiles(
            directory=os.path.dirname(os.path.realpath(__file__)) + "/frontend",
            html=True,
        ),
        name="base",
    )

    @api_app.get("/slicers")
    async def get_slicers():
        conn.send(("GET_SLICERS", ""))
        return conn.recv()

    @api_app.get("/slices")
    async def get_slices():
        conn.send(("GET_SLICES", ""))
        return conn.recv()

    @api_app.get("/table/{sli}")
    async def get_table(sli):
        conn.send(("GET_TABLE", sli))
        return Response(content=conn.recv())

    @api_app.get("/model_outputs/{res}/{model}")
    async def get_model_outputs(res, model):
        conn.send(("GET_MODEL_OUTPUTS", (res, model)))
        return conn.recv()

    @api_app.get("/metrics")
    async def get_metrics():
        conn.send(("GET_METRICS", ""))
        return conn.recv()

    @api_app.get("/data")
    async def get_data():
        conn.send(("GET_DATA", ""))
        return Response(content=conn.recv())

    @api_app.websocket("/results")
    async def results_websocket(websocket: WebSocket):
        await websocket.accept()
        previous_status = ""
        while True:
            await asyncio.sleep(1)
            conn.send(("GET_RESULTS", ""))
            res = conn.recv()
            if res[0] != previous_status:
                print("status: ", res[0])
                previous_status = res[0]
                await websocket.send_json({"status": res[0], "results": res[1]})

    uvicorn.run(app, host="localhost", port=8000)  # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeno']

package_data = \
{'': ['*'], 'zeno': ['frontend/*', 'frontend/build/*']}

install_requires = \
['fastapi>=0.75.0,<0.76.0',
 'modin>=0.13.2,<0.14.0',
 'pandas-stubs>=1.2.0,<2.0.0',
 'pandas>=1.4.0,<2.0.0',
 'pyarrow>=7.0.0,<8.0.0',
 'ray>=1.11.0,<2.0.0',
 'uvicorn>=0.17.5,<0.18.0',
 'watchdog>=2.1.6,<3.0.0',
 'websockets>=10.2,<11.0']

entry_points = \
{'console_scripts': ['zeno = zeno.runner:main']}

setup_kwargs = {
    'name': 'zenoml',
    'version': '0.0.1',
    'description': 'Behavioral Testing for Machine Learning',
    'long_description': '# Zeno - Behavioral testing of AI/ML\n\n![Github Actions CI tests](https://github.com/cabreraalex/zeno/actions/workflows/test.yml/badge.svg)\n[![code style black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![codecov](https://codecov.io/gh/cabreraalex/zeno/branch/main/graph/badge.svg?token=XPT8R98H8J)](https://codecov.io/gh/cabreraalex/zeno)\n\n## Quick Start\n\nInstall the Zeno package from PyPI:\n\n```\npip install zenoml\n```\n\n### [Follow the documentation to get started](https://cabreraalex.github.io/zeno/intro.html)\n\n## Development\n\n### Environment\n\nPlease install [`Poetry`](https://python-poetry.org/docs/master/#installing-with-the-official-installer) and use VSCode as your editor.\n\n### Install\n\nSuggest setting poetry to install the virtual env locally, which VSCode can use directly:\n\n`poetry config virtualenvs.in-project true`\n\n`poetry install`\n\n### Running\n\n`poetry run zeno`\n\n### Formatting and Linting\n\n`make`\n\n### Testing\n\n`make test`\n\n### Build Docs\n\n`make book`\n\n### Build\n\n`make build`\n\n### Publish\n\n`make publish`\n',
    'author': 'Ángel Alexander Cabrera',
    'author_email': 'alex.cabrera@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cabreraalex/zeno',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
from typing import Any, Callable, Union

import pandas as pd
import pyarrow as pa  # type: ignore
from watchdog.events import FileSystemEventHandler  # type: ignore


def get_arrow_bytes(df):
    df_arrow = pa.Table.from_pandas(df)
    buf = pa.BufferOutputStream()
    with pa.ipc.new_file(buf, df_arrow.schema) as writer:
        writer.write_table(df_arrow)
    return bytes(buf.getvalue())


# Used for preprocess and model outputs.
def cached_process(
    data_loader: Callable,
    df: pd.DataFrame,
    fn: Callable,
    cache,
    batch_size: int,
    id_column: str,
    data_path: str,
    transform: Union[Callable, None] = None,
):
    final_outputs: Any = [None] * df.shape[0]
    to_output = []
    outputs = []

    # Get all instances from cache possible
    for i, inst in enumerate(list(df[id_column])):
        if transform:
            inst = inst + "_T"
        if inst in cache:
            final_outputs[i] = cache[inst]
        else:
            to_output.append(i)

    if len(to_output) > 0:
        if len(to_output) < batch_size:
            data = data_loader(df.iloc[to_output], id_column, data_path)
            if transform:
                data = transform(data)
            outputs = fn(data)
        else:
            for i in range(0, len(to_output), batch_size):
                data = data_loader(
                    df.iloc[to_output[i : i + batch_size]], id_column, data_path
                )
                if transform:
                    data = transform(data)
                outputs.extend(fn(data))

        j = 0
        for i, inst in enumerate(final_outputs):
            if inst is None:
                final_outputs[i] = outputs[j]
                if transform:
                    cache[df.iloc[i][id_column] + "_T"] = final_outputs[i]
                else:
                    cache[df.iloc[i][id_column]] = final_outputs[i]

                j = j + 1

    return final_outputs


class TestFileUpdateHandler(FileSystemEventHandler):
    def __init__(self, files, callback):
        self.files = files
        self.callback = callback

    def on_modified(self, event):
        if os.path.abspath(event.src_path) in self.files:
            self.callback()



import asyncio
import os
import shelve
import sys
import threading
from importlib import util
from inspect import getmembers, getsource, isfunction, signature
from pathlib import Path
from typing import Callable, Dict, List, Union

import pandas as pd
from watchdog.observers import Observer  # type: ignore

from .util import cached_process, get_arrow_bytes, TestFileUpdateHandler


class Zeno(object):
    def __init__(
        self,
        metadata_path: Path,
        test_files: List[Path],
        models: List[Path],
        batch_size=16,
        id_column="id",
        label_column="label",
        data_path="",
        cache_path="",
    ):
        self.metadata_path = metadata_path
        self.test_files = test_files
        self.model_names = models
        self.batch_size = batch_size
        self.id_column = id_column
        self.label_column = label_column
        self.data_path = data_path
        self.cache_path = cache_path
        os.makedirs(self.cache_path, exist_ok=True)

        self.status: str = "Initializing"

        self.model_loader: Callable = None  # type: ignore
        self.data_loader: Callable = None  # type: ignore

        # Key is name of preprocess function
        self.preprocessors: Dict[str, Preprocessor] = {}
        # Key is name of slicer function
        self.slicers: Dict[str, Slicer] = {}
        # Key is name of transform function
        self.transforms: Dict[str, Transform] = {}
        # Key is name of metric function
        self.metrics: Dict[str, Metric] = {}
        # Key is name of slice
        self.slices: Dict[str, Slice] = {}
        # Key is result.id, hash of properties
        self.results: Dict[int, Result] = {}

        self._loop = None
        self.__initialize_watchdog()

        # Read metadata as Pandas for slicing
        self.metadata = pd.read_csv(metadata_path)

    def __initialize_watchdog(self):
        # Watch for updated test files.

        observer = Observer()
        event_handler = TestFileUpdateHandler(
            [os.path.abspath(t) for t in self.test_files], self.start_processing
        )
        folders_logged = []
        for test in self.test_files:
            folder = os.path.dirname(test)
            if folder not in folders_logged:
                folders_logged.append(folder)
                observer.schedule(event_handler, folder, recursive=True)
        observer.start()

    def start_processing(self):
        self.__parse_testing_files()
        self.__run_tests()

    def __parse_testing_files(self):
        self.data_loader = None  # type: ignore
        self.model_loader = None  # type: ignore
        self.slicers = {}
        self.metrics = {}
        self.transforms = {}

        for test_file in self.test_files:
            if test_file.is_dir():
                [
                    self.__parse_testing_file(f, test_file)
                    for f in list(test_file.rglob("*.py"))
                ]
            else:
                self.__parse_testing_file(test_file, test_file.parents[0])

        if not self.model_loader:
            print("ERROR: No model loader found")
            sys.exit(1)

        if not self.data_loader:
            print("ERROR: No data loader found")
            sys.exit(1)

    def __parse_testing_file(self, test_file: Path, base_path: Path):
        print(
            test_file,
            test_file.relative_to(base_path),
            test_file.relative_to(base_path).parts,
        )

        spec = util.spec_from_file_location("module.name", test_file)
        test_module = util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(test_module)  # type: ignore

        for func_name, func in getmembers(test_module):
            if isfunction(func):
                if hasattr(func, "load_model"):
                    if self.model_loader is None:
                        self.model_loader = func
                    else:
                        print("Multiple model loaders found, can only have one")
                        sys.exit(1)
                if hasattr(func, "load_data"):
                    if self.data_loader is None:
                        self.data_loader = func
                    else:
                        print("Multiple data loaders found, can only have one")
                        sys.exit(1)
                if hasattr(func, "preprocess"):
                    self.preprocessors[func_name] = Preprocessor(func_name, func)
                if hasattr(func, "slicer"):
                    path_names = list(test_file.relative_to(base_path).parts)
                    path_names[-1] = path_names[-1][0:-3]
                    self.slicers[func_name] = Slicer(
                        func_name,
                        func,
                        [*path_names, func_name],
                    )
                if hasattr(func, "transform"):
                    self.transforms[func_name] = Transform(func_name, func)
                if hasattr(func, "metric"):
                    self.metrics[func_name] = Metric(func_name, func)

    def __run_tests(self):
        if self._loop:
            self._loop.stop()
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._loop.run_forever, daemon=True)
        self._thread.start()
        self._loop.call_soon_threadsafe(asyncio.create_task, self.__run_test_pipeline())

    async def __run_test_pipeline(self):
        self.__preprocess_data()
        self.__slice_data()
        self.__calculate_metrics()

    def __preprocess_data(self):
        for name, preprocess in self.preprocessors.items():
            self.status = "Running preprocessor " + name
            cache = shelve.open(
                os.path.join(
                    self.cache_path,
                    (
                        ".model_preprocess_"
                        + str(self.metadata_path).replace("/", "_")
                        + name
                    ),
                ),
                writeback=True,
            )
            output = cached_process(
                self.data_loader,
                self.metadata,
                preprocess.func,
                cache,
                self.batch_size,
                self.id_column,
                self.data_path,
            )
            if not isinstance(output, list):
                print("ERROR: preprocess function must return a list")
                sys.exit(1)
            self.metadata[name] = output
        self.status = "Done preprocessing"

    def __slice_data(self):
        self.slices = {}

        for name, slicer in self.slicers.items():
            self.status = "Slicing data for " + name
            self.slices = {
                **self.slices,
                **slicer.slice_data(self.metadata, self.label_column),
            }
        self.status = "Done slicing"

    def __calculate_metrics(self):
        self.results = {}
        for i, model_name in enumerate(self.model_names):
            self.status = "Loading model {0}, {1}/{2}".format(
                model_name, str(i + 1), str(len(self.model_names))
            )

            model = self.model_loader(model_name)
            model_cache = self.__open_cache(".model_" + os.path.basename(model_name))

            for j, sli in enumerate(self.slices.values()):
                for k, metric in enumerate(sli.metrics):

                    transform = None
                    if type(metric) in (list, tuple):
                        transform = self.transforms[metric[0]]
                        metric = metric[1]
                    if transform is not None:
                        transform_name = transform.name
                        transform_fn = transform.func
                    else:
                        transform_name = ""
                        transform_fn = None

                    self.status = (
                        "Model {0}/{1}: Slice {2}/{3} ({4}),"
                        + " test {5}/{6} ({7}), {8} instances"
                    ).format(
                        str(i + 1),
                        str(len(self.model_names)),
                        str(j + 1),
                        str(len(self.slices.items())),
                        sli.name,
                        str(k + 1),
                        str(len(sli.metrics)),
                        metric,
                        str(sli.size),
                    )

                    if i == 0:
                        res = Result(
                            sli.name, transform_name, metric, sli.size, self.model_names
                        )
                    else:
                        res = self.results[
                            int(
                                hash("".join(sli.name) + transform_name + metric)
                                / 10000
                            )
                        ]

                    metric_func = self.metrics[metric].func

                    out = cached_process(
                        self.data_loader,
                        self.metadata.iloc[sli.sliced_indices],
                        model,
                        model_cache,
                        self.batch_size,
                        self.id_column,
                        self.data_path,
                        transform=transform_fn,
                    )

                    if len(signature(metric_func).parameters) == 3:
                        result = metric_func(
                            out,
                            self.metadata.iloc[sli.sliced_indices],
                            self.label_column,
                        )
                    else:
                        result = metric_func(
                            out,
                            self.metadata.iloc[sli.sliced_indices],
                        )

                    res.set_result(model_name, out, result)
                    self.results[res.id] = res
            model_cache.close()
        self.status = "Done"

    def get_table(self, sli):
        print(sli)
        df = self.metadata.iloc[self.slices[sli].sliced_indices]
        return get_arrow_bytes(df)

    def get_model_outputs(self, opts):
        res, model_hash = opts
        # TODO: figure out hash issue, truncates on frontend
        result = self.results[int(res)]
        return {
            "metric": result.model_metric_outputs[int(model_hash)],
            "output": result.model_outputs[int(model_hash)],
        }

    def get_metadata_path(self):
        return self.metadata_path

    def get_metrics(self):
        return self.metrics.values()

    def get_results(self):
        return self.results.values()

    def get_slicers(self):
        return self.slicers.values()

    def get_slices(self):
        return self.slices.values()

    def get_slice(self, sli):
        return self.slices[sli]

    def get_status(self):
        return self.status

    def __open_cache(self, name):
        return shelve.open(
            os.path.join(self.cache_path, name),
            writeback=True,
        )


class Slicer:
    def __init__(self, name: str, func: Callable, name_list: List[str]):
        self.name = name
        self.func = func
        self.name_list = name_list

        self.metrics = func.metrics  # type: ignore
        self.source = getsource(self.func)
        self.slices: List[str] = []

    def slice_data(self, metadata: pd.DataFrame, label_column: str):
        if len(signature(self.func).parameters) == 2:
            slicer_output = self.func(metadata, label_column)
        else:
            slicer_output = self.func(metadata)

        if isinstance(slicer_output, pd.DataFrame):
            slicer_output = slicer_output.index

        slicer_output = list(slicer_output)

        slices_to_return = {}
        # Can either be of the from [index list] or [(name, index list)..]
        if (
            len(slicer_output) > 0
            and isinstance(slicer_output[0], tuple)
            or isinstance(slicer_output[0], list)
        ):
            for output_slice in slicer_output:
                indices = list(output_slice[1])
                # Join slice names with backslash
                name_list = [*self.name_list, output_slice[0]]
                self.slices.append("".join(name_list))
                slices_to_return["".join(name_list)] = Slice(
                    [*self.name_list, output_slice[0]],
                    indices,
                    len(indices),
                    self.metrics,
                    self.name,
                )
        else:
            self.slices.append("".join(self.name_list))
            slices_to_return["".join(self.name_list)] = Slice(
                self.name_list,
                slicer_output,
                len(slicer_output),
                self.metrics,
                self.name,
            )

        return slices_to_return


class Preprocessor:
    def __init__(self, name: str, func: Callable):
        self.name = name
        self.func = func
        self.source = getsource(self.func)


class Slice:
    def __init__(
        self,
        name: List[str],
        sliced_indices: List[int],
        size: int,
        metrics: List[str],
        slicer: str,
    ):
        self.name = name
        self.sliced_indices = sliced_indices
        self.size = size
        self.metrics = metrics
        self.slicer = slicer


class Transform:
    def __init__(self, name: str, func: Callable):
        self.name = name
        self.func = func
        self.source = getsource(self.func)


class Metric:
    def __init__(self, name: str, func: Callable):
        self.name = name
        self.func = func
        self.source = getsource(self.func)


class Result:
    def __init__(
        self,
        sli: List[str],
        transform: str,
        metric: str,
        slice_size: int,
        model_names: List[Path],
    ):
        self.sli = sli
        self.transform = transform
        self.metric = metric
        self.slice_size = slice_size
        self.model_names = model_names

        # Keys are hash of model name
        self.model_outputs: Dict[int, list] = {}
        self.model_results: Dict[int, float] = {}
        self.model_metric_outputs: Dict[int, list] = {}

        self.id: int = int(
            hash("".join(self.sli) + self.transform + self.metric) / 10000
        )

    def set_result(self, model: Path, outputs: list, result: Union[pd.Series, list]):
        model_hash = int(hash(model) / 10000)
        self.model_outputs[model_hash] = outputs
        if isinstance(result, pd.Series):
            self.model_metric_outputs[model_hash] = result.astype(int).to_list()
        elif len(result) > 0 and isinstance(result[0], bool):
            self.model_metric_outputs[model_hash] = [
                1 if r is True else 0 for r in result
            ]

        self.model_results[model_hash] = (
            sum(self.model_metric_outputs[model_hash])
            / len(self.model_metric_outputs[model_hash])
            * 100
        )

    def __str__(self):
        return "Test {0} for slice {1}, transform {2}, size {3}".format(
            self.metric, self.sli, self.transform, self.slice_size
        )

    def __hash__(self):
        return self.id
